import re
import sys
import time
import numpy as np
import os
import warnings

warnings.filterwarnings('ignore', category=FutureWarning)
# from torch.utils.tensorboard import SummaryWriter
# from tensorboardX import SummaryWriter
from shutil import copytree
import torch
import copy
import shutil
from collections import defaultdict
from .utils import include_patterns, logger, check_type, beam_device, check_element_type, print_beam_hyperparameters
import pandas as pd
import torch.multiprocessing as mp
from .utils import setup, cleanup, set_seed, find_free_port, check_if_port_is_available, is_notebook, find_port, \
    pretty_format_number, as_numpy
import torch.distributed as dist
from functools import partial
from argparse import Namespace
from tensorboard.notebook import start as start_tensorboard
from ._version import __version__
import inspect

done = mp.Event()


def gen_hparams_string(e):
    tensorboard_hparams = pd.read_pickle(os.path.join(e, "hparams.pkl"))
    return '/'.join([f"{k}_{v}" for k, v in tensorboard_hparams.items()])


def path_depth(path):
    return len(os.path.normpath(path).split(os.sep))


def beam_algorithm_generator(experiment, Alg, Dataset=None, alg_args=None, alg_kwargs=None,
                             dataset_args=None, dataset_kwargs=None):

    if alg_args is None:
        alg_args = tuple()
    if alg_kwargs is None:
        alg_kwargs = dict()
    if dataset_args is None:
        dataset_args = tuple()
    if dataset_kwargs is None:
        dataset_kwargs = dict()

    if isinstance(Dataset, torch.utils.data.Dataset):
        dataset = Dataset
    elif Dataset is not None:
        dataset = Dataset(experiment.hparams, *dataset_args, **dataset_kwargs)
    else:
        dataset = None

    if type(Alg) == type:

        ars = inspect.getfullargspec(Alg)

        # don't pass dataset if the algorithm cannot handle it on initialization
        if 'dataset' in ars.args or ars.varargs is not None:
            alg = Alg(experiment.hparams, *alg_args, dataset=dataset, **alg_kwargs)
        else:
            alg = Alg(experiment.hparams, *alg_args, **alg_kwargs)
        # if a new algorithm is generated, we clean the tensorboard writer. If the reload option is True,
        # the algorithm will fix the epoch number s.t. tensorboard graphs will not overlap
        experiment.writer_cleanup()
    else:
        alg = Alg

    if alg.dataset is None and dataset is not None:
        alg.load_dataset(dataset)

    alg.experiment = experiment

    return alg


def default_runner(rank, world_size, experiment, algorithm_generator, *args, tensorboard_arguments=None, **kwargs):
    alg = algorithm_generator(*args, **kwargs)

    experiment.writer_control(enable=not (bool(rank)))
    results = {}

    try:
        for i, results in enumerate(iter(alg)):
            experiment.save_model_results(copy.deepcopy(results), alg, i,
                                          print_results=experiment.hparams.print_results,
                                          visualize_results=experiment.hparams.visualize_results,
                                          store_results=experiment.hparams.store_results, store_networks=experiment.hparams.store_networks,
                                          visualize_weights=experiment.hparams.visualize_weights,
                                          argv=tensorboard_arguments)

    except KeyboardInterrupt:

        logger.warning(f"KeyboardInterrupt: Training was interrupted, Worker terminates")
        if rank == 0:
            checkpoint_file = os.path.join(experiment.checkpoints_dir, f'checkpoint_{alg.epoch+1:06d}')
            alg.save_checkpoint(checkpoint_file)

    if world_size == 1:
        return alg, results


def run_worker(rank, world_size, results_queue, job, experiment, *args, **kwargs):

    logger.info(f"Worker: {rank + 1}/{world_size} is running...")

    if world_size > 1:
        setup(rank, world_size, port=experiment.hparams.mp_port)

    experiment.set_rank(rank, world_size)
    set_seed(seed=experiment.hparams.seed, constant=rank+1, increment=False, deterministic=experiment.hparams.deterministic)

    res = job(rank, world_size, experiment, *args, **kwargs)

    if world_size > 1:

        cleanup(rank, world_size)
        results_queue.put({'rank': rank, 'results': res})

        done.wait()

    else:
        return res


class Experiment(object):
    """
    Experiment name:
    <algorithm name>_<identifier>_exp_<number>_<time>


    Experiment number and overriding experiments

    These parameters are responsible for which experiment to load or to generate:
    the name of the experiment is <alg>_<identifier>_exp_<num>_<time>
    The possible configurations:
    reload = False, override = True: always overrides last experiment (default configuration)
    reload = False, override = False: always append experiment to the list (increment experiment num)
    reload = True, resume = -1: resume to the last experiment
    reload = True, resume = <n>: resume to the <n> experiment


    :param args:
    """

    def __init__(self, args, hpo=None, trial=None, print_hyperparameters=None, reload_iloc=-1,
                 reload_loc=None, reload_name=None):
        """

        @param args:
        @param hpo:
        @param trial:
        @param print_hyperparameters: If None, default behavior is to print hyperparameters only outside of jupyter notebooks
        """

        if print_hyperparameters is None:
            print_hyperparameters = not is_notebook()

        self.tensorboard_hparams = {}

        hparams = args.hparams
        self.vars_args = copy.copy(vars(args))
        for k, v in self.vars_args.items():
            param_type = check_type(v)
            if param_type.major == 'scalar' and param_type.element in ['bool', 'str', 'int', 'float'] and k in hparams:
                self.tensorboard_hparams[k] = v

        self.hparams = copy.copy(args)

        set_seed(seed=self.hparams.seed, constant=0, increment=False, deterministic=self.hparams.deterministic)

        # parameters
        self.start_time = time.time()
        self.exptime = time.strftime("%Y%m%d_%H%M%S", time.localtime())
        self.hparams.device = beam_device(self.hparams.device)

        self.base_dir = os.path.join(self.hparams.root_dir, self.hparams.project_name,
                                     self.hparams.algorithm, self.hparams.identifier)
        os.makedirs(self.base_dir, exist_ok=True)

        self.exp_name = None
        self.load_model = False

        pattern = re.compile("\A\d{4}_\d{8}_\d{6}\Z")
        exp_names = list(filter(lambda x: re.match(pattern, x) is not None, os.listdir(self.base_dir)))
        exp_indices = np.array([int(d.split('_')[0]) for d in exp_names])

        if self.hparams.reload:

            if type(self.hparams.resume) is str:
                if os.path.isdir(os.path.join(self.base_dir, self.hparams.resume)):
                    self.exp_name = self.hparams.resume
                    exp_num = int(self.exp_name.split('_')[0])
                    self.load_model = True

            elif self.hparams.resume >= 0:
                ind = np.nonzero(exp_indices == self.hparams.resume)[0]
                if len(ind):
                    self.exp_name = exp_names[ind[0]]
                    exp_num = self.hparams.resume
                    self.load_model = True

            else:
                if len(exp_indices):
                    ind = np.argmax(exp_indices)
                    self.exp_name = exp_names[ind]
                    exp_num = exp_indices[ind]
                    self.load_model = True

        else:

            if self.hparams.override and len(exp_indices):

                ind = np.argmax(exp_indices)
                self.exp_name = exp_names[ind]
                exp_num = exp_indices[ind]
            else:
                self.hparams.override = False

        if self.hparams.reload and not self.load_model:
            logger.warning(f"Did not find existing experiment to match your specifications: basedir={self.base_dir} resume={self.hparams.resume}")

        if self.exp_name is None:
            exp_num = np.max(exp_indices) + 1 if len(exp_indices) else 0
            self.exp_name = "%04d_%s" % (exp_num, self.exptime)

        # init experiment parameters
        self.root = os.path.join(self.base_dir, self.exp_name)

        # set dirs
        self.tensorboard_dir = os.path.join(self.root, 'tensorboard')
        self.checkpoints_dir = os.path.join(self.root, 'checkpoints')
        self.results_dir = os.path.join(self.root, 'results')
        self.code_dir = os.path.join(self.root, 'code')

        if self.load_model:
            logger.info(f"Resuming existing experiment (Beam version: {__version__})")

        else:

            if not self.hparams.override:
                logger.info(f"Creating new experiment (Beam version: {__version__})")

            else:
                logger.warning("Deleting old experiment")

                shutil.rmtree(self.root)
                self.exp_name = "%04d_%s" % (exp_num, self.exptime)
                self.root = os.path.join(self.base_dir, self.exp_name)

                # set dirs
                self.tensorboard_dir = os.path.join(self.root, 'tensorboard')
                self.checkpoints_dir = os.path.join(self.root, 'checkpoints')
                self.results_dir = os.path.join(self.root, 'results')
                self.code_dir = os.path.join(self.root, 'code')

            logger.info(f"Experiment directory is: {self.root}")

            os.makedirs(os.path.join(self.tensorboard_dir, 'logs'))
            os.makedirs(os.path.join(self.tensorboard_dir, 'hparams'))
            os.makedirs(self.checkpoints_dir)

            # make log dirs
            os.makedirs(os.path.join(self.results_dir))

            # copy code to dir

            if is_notebook():
                code_root_path = os.getcwd()
            else:
                code_root_path = sys.argv[0]

            copytree(os.path.dirname(os.path.realpath(code_root_path)), self.code_dir,
                     ignore=include_patterns('*.py', '*.md', '*.ipynb'))

            pd.to_pickle(self.vars_args, os.path.join(self.root, "args.pkl"))

        self.writer = None

        self.rank = 0
        self.world_size = args.parallel

        if self.world_size > 1:
            torch.multiprocessing.set_sharing_strategy('file_system')

        log_file = os.path.join(self.root, "experiment.log")
        logger.add(log_file, level='INFO', colorize=True,
                   format='<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>')

        print_beam_hyperparameters(args, debug_only=not print_hyperparameters)

        # replace zero split_dataset_seed to none (non-deterministic split) - if zero
        if self.hparams.split_dataset_seed == 0:
            self.hparams.split_dataset_seed = None

        # fill the batch size

        if self.hparams.batch_size_train is None:
            self.hparams.batch_size_train = self.hparams.batch_size

        if self.hparams.batch_size_eval is None:
            self.hparams.batch_size_eval = self.hparams.batch_size

        if self.hparams.batch_size is None:
            self.hparams.batch_size = self.hparams.batch_size_train

        # build the hyperparamter class which will be sent to the dataset and algorithm classes

        if self.load_model:
            self.hparams.reload_path = self.reload_checkpoint(iloc=reload_iloc, loc=reload_loc, name=reload_name)
        else:
            self.hparams.reload_path = None

        self.trial = trial

        self.hparams.hpo = hpo
        self.hparams.ddp = False
        self.hparams.rank = self.rank
        self.hparams.world_size = self.world_size

        if self.hparams.device.type == 'cuda':
            if self.hparams.device_list is not None:
                self.hparams.device_list = [beam_device(di) for di in self.hparams.device_list]
            else:
                self.hparams.device_list = [beam_device(di+self.hparams.device.index) for di in range(self.hparams.parallel)]

        pd.to_pickle(self.tensorboard_hparams, os.path.join(self.root, "hparams.pkl"))

    @staticmethod
    def reload_from_path(path, override_hparams=None, **argv):

        logger.info(f"Reload experiment from path: {path}")
        args = pd.read_pickle(os.path.join(path, "args.pkl"))
        args = Namespace(**args)
        args.override = False
        args.reload = True

        path, d = os.path.split(path)
        if not d:
            path, d = os.path.split(path)
        args.resume = d

        if override_hparams is not None:
            for k, v in override_hparams.items():
                setattr(args, k, v)

        return Experiment(args, **argv)

    def reload_checkpoint(self, alg=None, iloc=-1, loc=None, name=None):

        if name is not None:
            path = os.path.join(self.checkpoints_dir, name)

        else:

            checkpoints = os.listdir(self.checkpoints_dir)
            checkpoints = [c for c in checkpoints if c.split('_')[-1].isnumeric()]
            checkpoints_int = [int(c.split('_')[-1]) for c in checkpoints]

            if not(len(checkpoints)):
                logger.error(f"Directory of checkpoints does not contain valid checkpoint files")
                return

            checkpoints = pd.DataFrame({'name': checkpoints}, index=checkpoints_int)
            checkpoints = checkpoints.sort_index()

            if loc is not None:
                chp = checkpoints.loc[loc]['name']
                path = os.path.join(self.checkpoints_dir, chp)
            else:
                chp = checkpoints.iloc[iloc]['name']
                path = os.path.join(self.checkpoints_dir, chp)

        logger.info(f"Reload experiment from checkpoint: {path}")

        if alg is not None:
            alg.load_checkpoint(path)

        else:
            return path

    def set_rank(self, rank, world_size):

        self.rank = rank
        self.world_size = world_size

        self.hparams.rank = rank
        self.hparams.world_size = world_size

        self.hparams.ddp = self.world_size > 1
        self.hparams.enable_tqdm = self.hparams.enable_tqdm and (rank == 0)

        if self.hparams.device.type != 'cpu' and world_size > 1:
            self.hparams.device = beam_device(self.hparams.device_list[rank])

        logger.info(f'Worker {rank + 1} will be running on device={str(self.hparams.device)}')

    def writer_control(self, enable=True, networks=None, inputs=None):

        if enable and self.writer is None and self.hparams.tensorboard:
            from tensorboardX import SummaryWriter
            self.writer = SummaryWriter(log_dir=os.path.join(self.tensorboard_dir, 'logs'),
                                        comment=self.hparams.identifier)

        if networks is not None and enable and self.writer is not None:
            for k, net in networks.items():
                self.writer.add_graph(net, inputs[k])

    def save_model_results(self, results, algorithm, iteration, visualize_results='yes',
                           store_results='logscale', store_networks='logscale', print_results=True,
                           visualize_weights=False, argv=None):

        '''

        responsible for 4 actions:
        1. print results to stdout
        2. visualize results via tensorboard
        3. store results to pandas pickle objects
        4. save networks and optimizers

        logscale is active only on integer epochs in logscale (see x-axis in plt.semilogx)

        :param results:
        :param algorithm:
        :param visualize_results: takes yes|no|logscale.
        :param store_results: takes yes|no|logscale.
        :param store_networks: takes yes|no|logscale.
        :param print_results: whether to print the results to stdout when saving results to tensorboard.
        :return:
        '''

        epoch = algorithm.epoch
        if not self.rank:

            if print_results:
                logger.info('----------------------------------------------------------'
                            '---------------------------------------------------------------------')
                objective_str = ''
                if 'objective' in results and check_type(results['objective']).major == 'scalar':
                    objective_str = f"Current objective: {pretty_format_number(results['objective'])}"
                logger.info(f'Finished epoch {iteration+1}/{algorithm.n_epochs} (Total trained epochs {epoch}). '
                            f'{objective_str}')

            decade = int(np.log10(epoch) + 1)
            logscale = not (epoch - 1) % (10 ** (decade - 1))

            for subset, res in results.items():

                if store_results == 'yes' or store_results == 'logscale' and logscale:

                    os.makedirs(os.path.join(self.results_dir, subset), exist_ok=True)
                    torch.save(res, os.path.join(self.results_dir, subset, f'results_{epoch:06d}'))

                alg = algorithm if visualize_weights else None

            if visualize_results == 'yes' or visualize_results == 'logscale' and logscale:
                self.log_data(results, epoch, print_log=print_results, alg=alg, argv=argv)

            checkpoint_file = os.path.join(self.checkpoints_dir, f'checkpoint_{epoch:06d}')
            algorithm.save_checkpoint(checkpoint_file)
            if algorithm.best_state:
                checkpoint_file = os.path.join(self.checkpoints_dir, f'checkpoint_best')
                algorithm.save_checkpoint(checkpoint_file)

            if store_networks == 'no' or store_networks == 'logscale' and not logscale:
                try:
                    os.remove(os.path.join(self.checkpoints_dir, f'checkpoint_{epoch - 1:06d}'))
                except OSError:
                    pass

        if self.world_size > 1:
            dist.barrier()

    def log_data(self, results, n, print_log=True, alg=None, argv=None):

        def format_stat(k, v):
            format = f"{k if k != 'mean' else 'avg'}:{pretty_format_number(v)}".ljust(15)
            return format

        for subset, res in results.items():

            if subset == 'objective':
                continue

            def format(v):
                v_type = check_element_type(v)
                if v_type == 'int':
                    if v >= 1000:
                        return f"{float(v): .4}"
                    else:
                        return str(v)
                elif v_type == 'float':
                    return f"{v: .4}"
                else:
                    return v

            if print_log and 'stats' in res:
                logger.info(f'{subset}:')
                logger.info('| '.join([f"{k}: {format(v)} " for k, v in res['stats'].items()]))

            report = None
            if isinstance(res, dict):
                if 'scalar' in res:
                    report = res['scalar']
                else:
                    report = res

            if report is not None:

                for param, val in report.items():
                    stat = None
                    if type(val) is dict or type(val) is defaultdict:
                        for p, v in val.items():
                            val[p] = np.mean(v)
                    else:

                        v = [report[param]] if np.isscalar(report[param]) else report[param]
                        v = np.stack(v).flatten()

                        report[param] = np.mean(v)
                        if len(v) > 1 and np.var(v) > 0:
                            stat = pd.Series(v, dtype=np.float32).describe()
                        else:
                            v_type = check_type(v)
                            if v_type.major != 'scalar':
                                v = v[0]
                            v = int(v) if v_type.element == 'int' else float(v)
                            stat = {'val': v}

                    if print_log:
                        if not (type(report[param]) is dict or type(
                                report[param]) is defaultdict):
                            stat = '| '.join([format_stat(k, v) for k, v in dict(stat).items() if k != 'count'])

                            if len(param) > 11:
                                paramp = f'{param[:4]}...{param[-4:]}:'
                            else:
                                paramp = f'{param}:'

                            logger.info(f'{paramp: <12} | {stat}')

        if self.writer is None:
            return

        logger.info(f"Tensorboard results are stored to: {self.root}")
        defaults_argv = defaultdict(lambda: defaultdict(dict))
        if argv is not None:
            for log_type in argv:
                for k in argv[log_type]:
                    defaults_argv[log_type][k] = argv[log_type][k]

        if alg is not None:
            networks = alg.networks
            for net in networks:
                for name, param in networks[net].named_parameters():
                    try:
                        self.writer.add_histogram("weight_%s/%s" % (net, name), as_numpy(param), n,
                                                  bins='tensorflow')
                        self.writer.add_histogram("grad_%s/%s" % (net, name), as_numpy(param.grad), n,
                                                  bins='tensorflow')
                        if hasattr(param, 'intermediate'):
                            self.writer.add_histogram("iterm_%s/%s" % (net, name), as_numpy(param.intermediate),
                                                      n,
                                                      bins='tensorflow')
                    except:
                        pass
        metrics = {}
        for subset, res in results.items():
            if isinstance(res, dict) and subset != 'objective':
                for log_type in res:
                    if hasattr(self.writer, f'add_{log_type}'):
                        log_func = getattr(self.writer, f'add_{log_type}')
                        for param in res[log_type]:
                            if type(res[log_type][param]) is dict or type(res[log_type][param]) is defaultdict:
                                for p, v in res[log_type][param].items():
                                    log_func(f'{subset}_{param}/{p}', v, n, **defaults_argv[log_type][param])
                            elif type(res[log_type][param]) is list:
                                log_func(f'{subset}/{param}', *res[log_type][param], n, **defaults_argv[log_type][param])
                            else:
                                log_func(f'{subset}/{param}', res[log_type][param], n, **defaults_argv[log_type][param])
                                if log_type == 'scalar':
                                    metrics[f"{subset}/{param}"] = float(res[log_type][param])

        if len(metrics):
            self.writer.add_hparams(self.tensorboard_hparams, metrics, name=os.path.join('..', 'hparams'), global_step=n)
            # self.writer.add_hparams(self.tensorboard_hparams, metrics, run_name=os.path.join('..', 'hparams'))

    @staticmethod
    def _tensorboard(port=None, get_port_from_beam_port_range=True, base_dir=None, log_dirs=None, hparams=False):

        port = find_port(port=port, get_port_from_beam_port_range=get_port_from_beam_port_range)
        if port is None:
            return

        logger.info(f"Opening a tensorboard server on port: {port}")

        if hparams:
            command_argument = f"--bind_all --logdir {base_dir} --port {port}"
        else:
            command_argument = f"--bind_all --logdir_spec={log_dirs} --port {port}"
        start_tensorboard(command_argument)

    def normalize_experiment_path(self, path, level=0):

        normal_path = [self.hparams.root_dir, self.hparams.project_name,
                       self.hparams.algorithm, self.hparams.identifier]
        pd = path_depth(self.hparams.root_dir)

        return os.path.join(*normal_path[:len(normal_path)-pd-level], path)

    @staticmethod
    def open_tensorboard(root='', project=None, algorithm=None,
                         identifier=None, experiment=None, hparams=False, port=None,
                         get_port_from_beam_port_range=True):
        depth = 4
        filters = {'project': None, 'algorithm': None, 'identifier': None, 'experiment':None}
        if project is not None:
            path_type = check_type(project)
            if path_type.minor == 'list':
                filters['project'] = project
            else:
                filters['project'] = [project]
                path = os.path.join(root, project)
                if os.path.isdir(path):
                    root = path
                    depth = 3

        if algorithm is not None:
            path_type = check_type(algorithm)
            if path_type.minor == 'list':
                filters['algorithm'] = algorithm
            else:
                filters['algorithm'] = [algorithm]
                path = os.path.join(root, algorithm)
                if os.path.isdir(path):
                    root = path
                    depth = 2

        if identifier is not None:
            path_type = check_type(identifier)
            if path_type.minor == 'list':
                filters['identifier'] = identifier
            else:
                filters['identifier'] = [identifier]
                path = os.path.join(root, identifier)
                if os.path.isdir(path):
                    root = path
                    depth = 1
        if experiment is not None:
            path_type = check_type(experiment)
            if path_type.minor == 'list':
                filters['experiment'] = experiment
            else:
                filters['experiment'] = [experiment]
                path = os.path.join(root, experiment)
                if os.path.isdir(path):
                    root = path
                    depth = 0

        experiments = [d[0] for d in list(os.walk(root)) if (path_depth(d[0]) - path_depth(root)) == depth]
        experiments = [os.path.normpath(e) for e in experiments]

        if filters['project'] is not None:
            experiments = list(filter(lambda x: x.split(os.sep)[-4] in filters['project'], experiments))
        if filters['algorithm'] is not None:
            experiments = list(filter(lambda x: x.split(os.sep)[-3] in filters['algorithm'], experiments))
        if filters['identifier'] is not None:
            experiments = list(filter(lambda x: x.split(os.sep)[-2] in filters['identifier'], experiments))
        if filters['experiment'] is not None:
            experiments = list(filter(lambda x: x.split(os.sep)[-1] in filters['experiment'], experiments))

        print(experiments)

        names = ['/'.join(e.split(os.sep)[-3:]) for e in experiments]
        names = [f"{n}/{gen_hparams_string(e)}" for n, e in zip(names, experiments)]

        experiments = [os.path.join(e, 'tensorboard', 'logs') for e in experiments]
        log_dirs = ','.join([f"{n}:{e}" for n, e in zip(names, experiments)])

        Experiment._tensorboard(port=port, get_port_from_beam_port_range=get_port_from_beam_port_range,
                                base_dir=root, log_dirs=log_dirs, hparams=hparams)

    def tensorboard(self, port=None, add_all_of_same_identifier=False, add_all_of_same_algorithm=False,
                          add_all_of_same_project=False, more_experiments=None, more_identifiers=None,
                          more_algorithms=None, get_port_from_beam_port_range=True, hparams=False):

        suffix = 'hparams' if hparams else 'logs'

        if add_all_of_same_project:
            base_dir = os.path.join(self.hparams.root_dir, self.hparams.project_name)
            depth = 3
        elif add_all_of_same_algorithm:
            base_dir = os.path.join(self.hparams.root_dir, self.hparams.project_name, self.hparams.algorithm)
            depth = 2
        elif add_all_of_same_identifier:
            base_dir = os.path.join(self.hparams.root_dir, self.hparams.project_name, self.hparams.algorithm, self.hparams.identifier)
            depth = 1
        else:
            base_dir = self.root
            depth = 0

        experiments = [d[0] for d in list(os.walk(base_dir)) if (path_depth(d[0]) - path_depth(base_dir)) == depth]

        if more_experiments is not None:
            if hparams:
                logger.error("hparams visualization does not support adding additional experiments")
            if type(more_experiments) is str:
                more_experiments = [more_experiments]
                experiments = experiments + [self.normalize_experiment_path(e, level=0) for e in more_experiments]

        if more_identifiers is not None:
            if hparams:
                logger.error("hparams visualization does not support adding additional experiments")
            if type(more_identifiers) is str:
                more_identifiers = [more_identifiers]
                depth = 1
                for identifier in more_identifiers:
                    identifier = self.normalize_experiment_path(identifier, level=depth)
                    experiments = experiments + [d[0] for d in list(os.walk(identifier)) if (path_depth(d[0]) - path_depth(identifier)) == depth]

        if more_algorithms is not None:
            if hparams:
                logger.error("hparams visualization does not support adding additional experiments")
            if type(more_algorithms) is str:
                more_algorithms = [more_algorithms]
                depth = 2
                for algorithm in more_algorithms:
                    algorithm = self.normalize_experiment_path(algorithm, level=depth)
                    experiments = experiments + [d[0] for d in list(os.walk(algorithm)) if (path_depth(d[0]) - path_depth(algorithm)) == depth]

        experiments = [os.path.normpath(e) for e in experiments]
        names = ['/'.join(e.split(os.sep)[-3:]) for e in experiments]
        names = [f"{n}/{gen_hparams_string(e)}" for n, e in zip(names, experiments)]

        experiments = [os.path.join(e, 'tensorboard', suffix) for e in experiments]
        log_dirs = ','.join([f"{n}:{e}" for n, e in zip(names, experiments)])

        self._tensorboard(port=port, get_port_from_beam_port_range=get_port_from_beam_port_range,
                          base_dir=base_dir, log_dirs=log_dirs, hparams=hparams)

    def algorithm_generator(self, Alg, Dataset=None, alg_args=None, alg_kwargs=None,
                             dataset_args=None, dataset_kwargs=None):
        return beam_algorithm_generator(self, Alg=Alg, Dataset=Dataset, alg_args=alg_args, alg_kwargs=alg_kwargs,
                             dataset_args=dataset_args, dataset_kwargs=dataset_kwargs)

    def fit(self, Alg=None, Dataset=None, *args, algorithm_generator=None, return_results=False, reload_results=False,
            tensorboard_arguments=None, alg_args=None, alg_kwargs=None, dataset_args=None,
            dataset_kwargs=None, **kwargs):

        if algorithm_generator is None:
            ag = partial(beam_algorithm_generator, Alg=Alg, Dataset=Dataset, alg_args=alg_args, alg_kwargs=alg_kwargs,
                                 dataset_args=dataset_args, dataset_kwargs=dataset_kwargs)
        else:

            if Alg is not None:
                ag = partial(algorithm_generator, Alg=Alg, Dataset=Dataset, alg_args=alg_args, alg_kwargs=alg_kwargs,
                                     dataset_args=dataset_args, dataset_kwargs=dataset_kwargs)
            else:
                ag = algorithm_generator

        return self(ag, *args, return_results=return_results, reload_results=reload_results,
                    tensorboard_arguments=tensorboard_arguments, **kwargs)

    def __call__(self, algorithm_generator, *args, return_results=False, reload_results=False,
                  tensorboard_arguments=None, **kwargs):

        try:
            res = self.run(default_runner, *(algorithm_generator, self, *args),
                           tensorboard_arguments=tensorboard_arguments, **kwargs)

        except KeyboardInterrupt:

            res = None
            logger.warning(f"KeyboardInterrupt: Training was interrupted, reloads last checkpoint")

        if res is None or self.world_size > 1:
            alg = algorithm_generator(self, *args, **kwargs)
            self.reload_checkpoint(alg)

            if reload_results:
                results = {}
                for subset in os.listdir(alg.results_dir):

                    res = os.listdir(os.path.join(alg.results_dir, subset))
                    res = pd.DataFrame({'name': res, 'index': [int(c.split('_')[-1]) for c in res]})
                    res = res.sort_values('index')

                    res = res.iloc['name']
                    path = os.path.join(alg.results_dir, subset, res)

                    results[subset] = path

                if reload_results:
                    results = {subset: pd.read_pickle(path) for subset, path in results.items()}

        else:
            alg, results = res

        if return_results:
            return alg, results
        else:
            return alg

    def run(self, job, *args, **kwargs):

        arguments = (job, self, *args)

        def _run(demo_fn, world_size):

            ctx = mp.get_context('spawn')
            results_queue = ctx.Queue()
            for rank in range(world_size):
                ctx.Process(target=demo_fn, args=(rank, world_size, results_queue, *arguments),
                            kwargs=kwargs).start()

            res = []
            for rank in range(world_size):
                res.append(results_queue.get())

            done.set()

            return res

        if self.world_size > 1:
            logger.info(f'Initializing {self.world_size} parallel workers')
            logger.warning(f"Caution: Sometimes DDP experiments can fail due to a bad configuration. "
                           f"Specifically, if in_place error set --no-broadcast-buffer flag and for subgraph issues"
                           f"set --find-unused-parameters")

            if self.hparams.mp_port == 'random' or check_if_port_is_available(self.hparams.mp_port):
                self.hparams.mp_port = find_free_port()

            logger.info(f'Multiprocessing port is: {self.hparams.mp_port}')

            return _run(run_worker, self.world_size)
        else:
            logger.info(f'Single worker mode')
            return run_worker(0, 1, None, *arguments, **kwargs)

    def writer_cleanup(self):

        if self.writer is not None:
            self.writer.close()
            self.writer = None



import torch
import pandas as pd
import warnings
from collections import namedtuple
import numpy as np
from .utils import check_type, slice_to_index, as_tensor, is_boolean, as_numpy


class Iloc(object):

    def __init__(self, pointer):
        self.pointer = pointer

    def __getitem__(self, ind):
        return self.pointer._iloc(ind)


class Loc(object):

    def __init__(self, pointer):
        self.pointer = pointer

    def __getitem__(self, ind):
        return self.pointer._loc(ind)


class DataTensor(object):
    def __init__(self, data, columns=None, index=None, requires_grad=False, device=None, series=False, **kwargs):
        super().__init__(**kwargs)

        self.series = series

        if isinstance(data, pd.DataFrame):
            columns = data.columns
            index = data.index
            data = data.values
        elif isinstance(data, dict):
            columns = list(data.keys())
            data = torch.stack([data[c] for c in columns], dim=1)

        if not isinstance(data, torch.Tensor):
            data = as_tensor(data, **kwargs)

        self.device = data.device if device is None else device

        data = data.to(self.device)

        if requires_grad and not data.requires_grad:
            data.requires_grad_()

        assert len(data.shape) == 2, "DataTensor must be two-dimensional"
        n_rows, n_columns = data.shape

        index_type = check_type(index)
        if index is None:
            index = torch.arange(n_rows)
            self.index_map = None
            self.mapping_method = 'simple'
        else:
            if index_type.minor == 'tensor':
                index = as_numpy(index)
            elif index_type.major == 'scalar':
                index = [index]

            self.index_map = pd.Series(index=index, data=np.arange(len(index)))
            self.mapping_method = 'series'

        columns_type = check_type(columns)
        if columns is None:

            if data.shape[1] == 1:
                columns = ['']
                self.columns_format = 'str'

            else:
                columns = [int(i) for i in torch.arange(n_columns)]
                self.columns_format = 'int'

        elif columns_type.major == 'array' and columns_type.element == 'int':

            columns = [int(i) for i in columns]
            self.columns_format = 'int'

        elif columns_type.major == 'array':

            columns = [str(i) for i in columns]
            self.columns_format = 'str'

        else:
            raise ValueError

        self.columns_map = {str(k): i for i, k in enumerate(columns)}
        assert len(columns) == n_columns, "Number of keys must be equal to the tensor 2nd dim"

        self.index = index
        self.data = data
        self.columns = columns

        self.iloc = Iloc(self)
        self.loc = Loc(self)

    def __eq__(self, other):
        return self.values.__eq__(other)

    def __ge__(self, other):
        return self.values.__ge__(other)

    def __ne__(self, other):
        return self.values.__ne__(other)

    def __lt__(self, other):
        return self.values.__lt__(other)

    def __gt__(self, other):
        return self.values.__gt__(other)

    def __le__(self, other):
        return self.values.__le__(other)

    def __len__(self):
        return len(self.index)

    def inverse_columns_map(self, columns):

        cast = int if self.columns_format == 'int' else str

        if check_type(columns).major == 'scalar':
            columns = self.columns_map[cast(columns)]
        else:
            columns = [self.columns_map[cast(i)] for i in columns]

        return columns

    def inverse_map(self, ind):

        ind = slice_to_index(ind, l=len(self), sliced=self.index)

        index_type = check_type(ind)

        if self.mapping_method == 'simple':
            pass
        elif self.mapping_method == 'series':
            if index_type.minor == 'tensor':
                ind = as_numpy(ind)
            ind = as_tensor(self.index_map[ind].values, return_vector=True)
        else:
            return NotImplementedError

        return ind

    def apply(self, func, dim=0):

        def remove_dt(d):
            if type(d) is DataTensor:
                return d.sort_values
            return d

        if dim == 1:
            data = torch.concat([remove_dt(func(DataTensor(di.unsqueeze(0), columns=self.columns)))
                                        for di in self.data], dim=0)
        elif dim == 0:
            data = torch.concat([remove_dt(func(DataTensor(di.unsqueeze(0), index=self.index)) )
                                           for di in self.data.T], dim=1)

        return DataTensor(data, columns=self.columns, index=self.index)

    def save(self, path):

        index = None if self.mapping_method == 'simple' else self.index

        state = {'data': self.data, 'index': index, 'columns': self.columns,
                 'requires_grad': self.data.requires_grad, 'device': self.data.device, 'series': self.series}
        torch.save(state, path)

    @staticmethod
    def load(path, map_location=None):

        state = torch.load(path)
        device = map_location if map_location is not None else state['device']
        return DataTensor(state['data'], index=state['index'], device=device, columns=state['columns'],
                          series=state['series'], requires_grad=state['requires_grad'])

    @property
    def values(self):

        data = self.data
        if self.series:
            data = data.squeeze(1)

        return data

    def _iloc(self, ind):

        ind = slice_to_index(ind, l=self.data.shape[0])
        index_type = check_type(ind)

        if index_type.major == 'scalar':
            ind = [ind]

        index = self.inverse_map(ind)
        data = self.data[ind]

        return DataTensor(data, columns=self.columns, index=index)

    def to(self, device):
        self.data = self.data.to(device)

        return self

    def __repr__(self):

        if isinstance(self.index, torch.Tensor):
            index = as_numpy(self.index.data)
        else:
            index = self.index

        repr_data = repr(pd.DataFrame(data=as_numpy(self.data.detach()),
                                      columns=self.columns, index=index))

        inf = f'DataTensor:\ndevice:\t\t{self.data.device}\nrequires_grad:\t{self.data.requires_grad}'

        if self.data.requires_grad and self.data.grad_fn is not None:
            grad_info = f'\ngrad_fn:\t{self.data.grad_fn.name()}'
        else:
            grad_info = ''

        return f'{repr_data}\n\n{inf}{grad_info}'

    def __setitem__(self, ind, data):

        if type(data) is DataTensor:
            data = data.data

        if type(ind) is tuple:

            index = ind[0]
            columns = ind[1]

            ind_index = self.inverse_map(index)
            ind_columns = self.inverse_columns_map(columns)

            self.data[ind_index, ind_columns] = data
            return

        else:

            columns = ind

            existing_columns = set(self.columns).difference(columns)
            new_columns = set(columns).difference(self.columns)

            assert not len(existing_columns) * len(new_columns), "Cannot assign new and existing columns in a single operations"

            if len(existing_columns):

                ind_columns = self.inverse_columns_map(columns)
                self.data[:, ind_columns] = data
                return

            if len(existing_columns):

                if check_type(columns).major == 'scalar':

                    data = data.unsqueeze(1)
                    columns = [columns]

                data = torch.cat([self.data, data], dim=1)
                columns = self.columns + columns

                self.__init__(data, columns=columns, index=self.index)

        raise ValueError

    def sort_index(self, ascending=True):

        sorted_index = torch.sort(self.index, descending=not ascending).values
        return self.loc[sorted_index]

    def _loc(self, ind):

        series = False
        if type(ind) is tuple:

            index = ind[0]
            columns = ind[1]

            ind_columns = self.inverse_columns_map(columns)
            if check_type(ind_columns).major == 'scalar':
                ind_columns = [int(ind_columns)]
                columns = [columns]
                series = True

        else:

            index = ind
            columns = self.columns
            ind_columns = slice(None)

        index = slice_to_index(index, l=len(self), sliced=self.index)
        if check_type(index).major == 'scalar':
            index = [index]

        ind_index = self.inverse_map(index)
        data = self.data[ind_index][slice(None), ind_columns]

        return DataTensor(data, columns=columns, index=index, series=series)

    def __getitem__(self, ind):

        series = False

        if (len(ind) == self.data.shape[0]) and is_boolean(ind):

            if len(ind) == 1:
                ind = torch.where(ind)[0]

            data = self.data[ind]
            index = self.index[ind]

            return DataTensor(data, columns=self.columns, index=index)

        columns = ind
        ind_columns = self.inverse_columns_map(columns)
        if check_type(ind_columns).major == 'scalar':
            ind_columns = [int(ind_columns)]
            columns = [columns]
            series = True

        data = self.data[slice(None), ind_columns]
        return DataTensor(data, columns=columns, index=self.index, series=series)


prototype = torch.Tensor([0])


def decorator(f_str):
    def apply(x, *args, **kargs):

        f = getattr(x.data, f_str)

        args = list(args)
        for i, a in enumerate(args):
            if type(a) is DataTensor :
                args[i] = a.data

        for k, v in kargs.items():
            if type(v) is DataTensor:
                kargs[k] = v.data

        r = f(*args, **kargs)
        if 'return_types' in str(type(r)):
            data = r.values
        else:
            data = r

        if isinstance(data, torch.Tensor):

            if len(data.shape) == 2:
                n_rows, n_columns = data.shape

            elif len(data.shape) == 1 and len(x.index) != len(x.columns):

                warnings.warn("Warning: Trying to infer columns or index dimensions from the function output")

                if len(x.columns) == len(data):
                    n_columns = len(x.columns)
                    data = data.unsqueeze(0)
                    n_rows = 1

                elif len(x.index) == len(data):
                    n_rows = len(x.index)
                    data = data.unsqueeze(1)
                    n_columns = 1

                else:
                    return r

            else:
                return r

            index = x.index if n_rows == len(x.index) else [f_str]
            columns = x.columns if n_columns == len(x.columns) else [f_str]

            if index is not None or columns is not None:
                data = DataTensor(data, columns=columns, index=index)
                if 'return_types' in str(type(r)):

                    ReturnType = namedtuple(f_str, ['values', 'indices'])
                    r = ReturnType(data, r.indices)

                else:
                    r = data
        return r

    return apply

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    for p in dir(prototype):
        try:
            f = getattr(prototype, p)
            if callable(f) and p not in dir(DataTensor):
                setattr(DataTensor, p, decorator(p))
        except RuntimeError:
            pass
        except TypeError:
            pass






import torch
from .utils import check_type, slice_to_index, as_tensor, beam_device
import pandas as pd
from .data_tensor import DataTensor


class PackedFolds(object):

    def __init__(self, data, index=None, names=None, fold=None, fold_index=None, device=None,
                 sort_index=False, quick_getitem=True):

        device = beam_device(device)
        self.quick_getitem = quick_getitem
        self.names_dict = None
        self.names = None
        self.sampling_method = 'folds'

        if names is not None:
            self.names = names
            self.names_dict = {n: i for i, n in enumerate(self.names)}

        data_type = check_type(data)

        if data_type.minor == 'list':
            self.data = [as_tensor(v, device=device) for v in data]

        elif data_type.minor == 'dict':
            if names is None:
                self.names = list(data.keys())
                self.names_dict = {n: i for i, n in enumerate(self.names)}

            self.data = [as_tensor(data[k], device=device) for k in self.names]

        elif data_type.major == 'array':
            self.data = as_tensor(data, device=device)
            self.sampling_method = 'no_folds'
        else:
            raise ValueError("data should be either dict/list/array")

        fold_type = check_type(fold)

        if fold_type.element == 'str':
            fold, names_map = pd.factorize(fold)

            if self.names_dict is not None:
                assert all([i == self.names_dict[n] for i, n in enumerate(names_map)]), "fold and data maps must match"

            else:
                self.names = list(names_map)
                self.names_dict = {n: i for i, n in enumerate(self.names)}

            fold = as_tensor(fold)

            if self.sampling_method == 'no_folds':
                self.sampling_method = 'foldable'

        elif fold_type.element == 'int':
            fold = as_tensor(fold)

            if self.sampling_method == 'no_folds':
                self.sampling_method = 'foldable'

        elif fold is None:
            if self.sampling_method == 'no_folds':
                assert len(names) == 1, "this is the single fold case"
                fold = torch.zeros(len(self.data), dtype=torch.int64)
                self.sampling_method = 'offset'
            else:
                fold = torch.cat([i * torch.ones(len(d), dtype=torch.int64) for i, d in enumerate(self.data)])

        else:
            raise ValueError

        if data_type.minor in ['dict', 'list']:
            lengths = torch.LongTensor([len(di) for di in self.data])
        else:
            lengths = torch.bincount(fold)

        if fold_index is not None:
            fold_index = as_tensor(fold_index)
        else:
            fold_index = torch.cat([torch.arange(l) for l in lengths])

        # update names
        if self.names is None:
            if data_type.minor in ['dict', 'list']:
                self.names = list(range(len(self.data)))
            else:
                self.names = torch.sort(torch.unique(fold)).values

            self.names_dict = {n: i for i, n in enumerate(self.names)}

        # merge data if possible for faster slicing
        if data_type.minor in ['dict', 'list']:
            dtype = [di.dtype for di in self.data]
            shape = [di.shape[1:] for di in self.data]
            if all([d == dtype[0] for d in dtype]) and all([s == shape[0] for s in shape]):
                self.data = torch.cat(self.data)
                self.sampling_method = 'offset'

        index_type = check_type(index)

        if index_type.minor == 'list':
            index = torch.concat([as_tensor(v, return_vector=True) for v in index])

        elif index_type.minor == 'dict':
            index = torch.concat([as_tensor(index[k], return_vector=True) for k in self.names])

        elif index_type.major == 'array':
            index = as_tensor(index)

        elif index is None:

            if self.sampling_method in ['offset', 'foldable']:
                index = torch.arange(len(self.data))
                self.sampling_method = 'index'
            else:
                index = torch.arange(sum([len(d) for d in self.data]))

        else:
            raise ValueError

        cumsum = torch.cumsum(lengths, dim=0)
        offset = cumsum - lengths
        offset = offset[fold] + fold_index

        info = {'fold': fold, 'fold_index': fold_index, 'offset': offset}

        if self.sampling_method == 'index':
            index = None

        self.info = DataTensor(info, index=index, device=device)

        if sort_index:
            self.sort_index()

    def sort_index(self, ascending=True):

        self.info = self.info.sort_index(ascending=ascending)

        return self

    def __len__(self):
        return len(self.info)

    def get_fold(self, name):

        fold = self.names_dict[name]
        info = self.info[self.info['fold'] == fold]
        index = info.index

        if self.sampling_method == 'folds':
            data = self.data[fold]
        elif self.sampling_method == 'index':
            data = self.data[index]
        elif self.sampling_method == 'offset':
            data = self.data[info['offset'].values]
        else:
            raise Exception(f"Sampling method unsupported: {self.sampling_method}")

        return PackedFolds(data=data, index=index, names=[name], device=self.device)

    def apply(self, functions):

        functions_type = check_type(functions)

        if functions_type.minor == 'list':
            data = [f(d) for d, f in zip(self.data, functions)]

        elif functions_type.minor == 'dict':
            data = [f(self.get_fold(k)) for k, f in functions.items()]
        else:
            raise ValueError

        return PackedFolds(data=data, index=self.index, names=self.names, device=self.device)

    @property
    def index(self):
        return self.info.index

    @property
    def fold(self):
        return self.info['fold'].values

    @property
    def fold_index(self):
        return self.info['fold_index'].values

    @property
    def offset(self):
        return self.info['offset'].values

    @property
    def shape(self):

        if self.sampling_method == 'fold':
            shape = {k: d.shape for k, d in zip(self.names, self.data)}
        else:
            shape = self.data.shape
        return shape

    @property
    def values(self):

        if self.sampling_method == 'fold':
            data = torch.cat(self.data)
        else:
            data = self.data

        return data[self.info['offset'].values]

    @property
    def tag(self):
        if len(self.names) == 1:
            return self.names[0]
        return 'hetrogenous'

    @property
    def device(self):
        if self.sampling_method == 'folds':
            device = self.data[0].device
        else:
            device = self.data.device

        return device

    def to(self, device):

        if self.sampling_method == 'folds':
            self.data = [di.to(device) for di in self.data]
        else:
            self.data = self.data.to(device)

        self.info = self.info.to(device)

        return self

    def __repr__(self):
        if self.sampling_method == 'folds':
            data = {k: self.data[self.names_dict[k]] for k in self.names}
        else:
            data = {k: self.get_fold(k).data for k in self.names}

        if len(data) == 1:
            data = next(iter(data.values()))

        return repr(data)

    def __getitem__(self, ind):

        ind_type = check_type(ind, check_minor=False)
        if ind_type.major == 'scalar' and ind_type.element == 'str':
            return self.get_fold(ind)
        if self.sampling_method == 'index' and self.quick_getitem:
            return self.data[ind]

        if type(ind) is tuple:
            ind_rest = ind[1:]
            ind = ind[0]
        else:
            ind_rest = tuple()

        ind = slice_to_index(ind, l=len(self), sliced=self.index)
        if ind_type.major == 'scalar':
            ind = [ind]

        info = self.info.loc[ind]

        if self.sampling_method == 'folds':
            fold, fold_index = info[['fold', 'fold_index']].values.T

            uq = torch.sort(torch.unique(fold)).values
            names = [self.names[i] for i in uq]

            if len(uq) == 1:

                data = self.data[uq[0]].__getitem__((fold_index, *ind_rest))
                if len(ind) == 1:
                    return data[0]

                return PackedFolds(data=[data], names=names, index=ind, device=self.device)

            fold_index = [fold_index[fold == i] for i in uq]
            data = [self.data[i].__getitem__((j, *ind_rest)) for i, j in zip(uq, fold_index)]

            fold = None

        else:

            fold, offset = info[['fold', 'offset']].values.T
            index = info.index
            ind = index if self.sampling_method == 'index' else offset

            data = self.data.__getitem__((ind, *ind_rest))

            uq = torch.sort(torch.unique(fold)).values
            names = [self.names[i] for i in uq]

            if self.sampling_method == 'index':
                ind = None
            else:
                ind = index

        return PackedFolds(data=data, names=names, index=ind, fold=fold, device=self.device)


import copy

import torch
from torch import nn
from collections import defaultdict
import numpy as np
import math
from .utils import slice_to_index, logger, hash_tensor
from .optim import BeamOptimizer
from functools import partial
import random


class PackedSet(object):

    def __init__(self, data, length=None, device=None):

        if length is None:
            self.data = torch.cat(data, dim=0)
            self.length = torch.LongTensor([0] + [len(x) for x in data], device=self.data.device)
        else:
            self.data = data
            self.length = torch.LongTensor([0] + list(length), device=self.data.device)

        self._offset = self.length.cumsum(dim=0)
        self.length = self.length[1:]
        self.offset = self._offset[:-1]
        self.index = torch.arange(len(self.offset), device=self.data.device)

        if device is not None:
            self.to(device)

    def __len__(self):
        return len(self.offset)

    def clone(self):
        return PackedSet(self.data.clone(), self.length.clone())

    def to(self, device):
        self.data = self.data.to(device)
        self._offset = self._offset.to(device)
        self.length = self.length.to(device)
        self.offset = self.offset.to(device)
        self.index = self.index.to(device)
        return self

    def aggregate(self, func):
        return torch.stack([func(self.data[self._offset[i]:self._offset[i + 1]]) for i in range(len(self))])

    def __getitem__(self, index):

        index = slice_to_index(index, l=len(self))
        if isinstance(index, np.ndarray):
            if index.dtype == np.dtype('bool'):
                index = torch.BoolTensor(index)
            else:
                index = torch.LongTensor(index)
        elif type(index) is int:
            index = torch.scalar_tensor(index, dtype=torch.int64)
        if isinstance(index, torch.Tensor):
            if index.dtype == torch.bool:
                index = self.index[index]
            shape = index.shape
            if len(shape) == 0:
                return self.data[self._offset[index]:self._offset[index + 1]]
            elif len(shape) == 1:
                return PackedSet([self.data[self._offset[i]:self._offset[i + 1]] for i in index])
            else:
                raise NotImplementedError

        elif type(index) is tuple:
            assert len(index) == 2
            a, b = index
            return self.data[b + self._offset[a]]

        else:
            raise NotImplementedError

    def __repr__(self):
        return repr(self.data)


class PositionalHarmonicExpansion(object):

    def __init__(self, xs, xe, delta=2, rank='cuda', n=100):

        self.xs = xs
        self.xe = xe

        delta = delta / (xe - xs)
        i = torch.arange(n+1, device=rank)[None, :]
        self.sigma = 2 * ((1 / delta) ** (i / n))

    def transform(self, x):

        x = x.unsqueeze(-1)

        sin = torch.sin(2 * np.pi / self.sigma * x)
        cos = torch.cos(2 * np.pi / self.sigma * x)

        out = torch.cat([sin, cos], dim=-1)
        return out


class LinearNet(nn.Module):

    def __init__(self, l_in, l_h=256, l_out=1, n_l=2, bias=True, activation='ReLU'):
        super().__init__()

        activation = getattr(nn, activation)
        self.lin = nn.Sequential(*([nn.Linear(l_in, l_h, bias=bias), activation()] if n_l > 1 else []),
                                 *sum([[nn.Linear(l_h, l_h, bias=bias), activation()] for _ in range(max(n_l - 2, 0))],
                                      []),
                                 nn.Linear(l_h if n_l > 1 else l_in, l_out, bias=bias))

    def forward(self, x):

        y = self.lin(x)
        return y.squeeze(1)


class RuleLayer(nn.Module):

    def __init__(self, n_rules, e_dim_in, e_dim_out, bias=True, pos_enc=None, dropout=0.0):
        super().__init__()

        self.query = nn.Parameter(torch.empty((n_rules, e_dim_out)))
        nn.init.kaiming_uniform_(self.query, a=math.sqrt(5))

        self.key = nn.Linear(e_dim_in, e_dim_out, bias=bias)
        self.value = nn.Linear(e_dim_in, e_dim_out, bias=bias)
        self.e_dim_out = e_dim_out
        self.sparsemax = nn.Softmax(dim=1)
        self.tau = 1.

        if pos_enc is None:
            self.pos_enc = nn.Parameter(torch.empty((e_dim_in, e_dim_out)))
            nn.init.kaiming_uniform_(self.pos_enc, a=math.sqrt(5))

    def forward(self, x):
        b, nf, ne = x.shape

        pos = self.pos_enc[:nf].unsqueeze(0).repeat(b, 1, 1)
        x = x + pos

        k = self.key(x)
        v = self.value(x)
        q = self.query

        a = k @ q.T / math.sqrt(self.e_dim_out)
        a_prob = self.sparsemax(a / self.tau).transpose(1, 2)

        r = a_prob @ v

        return r, a_prob


class MHRuleLayer(nn.Module):

    def __init__(self, n_rules, n_features, e_dim_out, bias=True, pos_enc=None, dropout=0.0, n_head=8,
                 static_attention=False):
        super().__init__()

        self.query = nn.Parameter(torch.empty((n_rules, e_dim_out)))
        nn.init.kaiming_uniform_(self.query, a=math.sqrt(5))

        self.key = nn.Linear(e_dim_out, e_dim_out, bias=bias)
        self.value = nn.Linear(e_dim_out, e_dim_out, bias=bias)
        self.out = nn.Linear(e_dim_out, e_dim_out, bias=bias)
        self.e_dim_out = e_dim_out
        #         self.sparsemax = Sparsemax(dim=2)
        self.sparsemax = nn.Softmax(dim=2)
        self.tau = 1.
        self.n_head = n_head
        self.n_rules = n_rules

        if pos_enc is None:
            self.pos_enc = nn.Parameter(torch.empty((n_features, e_dim_out)))
            nn.init.kaiming_uniform_(self.pos_enc, a=math.sqrt(5))

    def forward(self, x):
        b, nf, ne = x.shape
        pos = self.pos_enc.unsqueeze(0).repeat(b, 1, 1)

        x = x + pos
        v = self.value(x)

        k = self.key(x)
        b, nf, ne = k.shape

        k = k.view(b, nf, self.n_head, ne // self.n_head).transpose(1, 2).reshape(b, self.n_head, nf, ne // self.n_head)
        v = v.view(b, nf, self.n_head, ne // self.n_head).transpose(1, 2).reshape(b, self.n_head, nf, ne // self.n_head)

        q = self.query.view(1, self.n_rules, self.n_head, ne // self.n_head).transpose(1, 2)

        a = k @ q.transpose(2, 3) / math.sqrt(self.e_dim_out // self.n_head)
        a_prob = self.sparsemax(a / self.tau).transpose(2, 3)

        r = (a_prob @ v).transpose(1, 2).reshape(b, self.n_rules, self.e_dim_out)
        r = self.out(r)

        return r, a_prob


class LazyLinearRegression(nn.Module):

    def __init__(self, quantiles, lr=0.001):
        super().__init__()

        self.lr = lr
        self.quantiles = quantiles
        self.register_buffer('a', None)
        self.register_buffer('b', None)

    def forward(self, e):

        _, e_dim = e.shape
        e = e.view(-1, self.quantiles + 2, e_dim)[:, 1:]

        x = torch.arange(self.quantiles + 1, device=e.device) / self.quantiles - 0.5
        x = x.reshape(1, -1, 1)

        if self.b is None:
            self.b = e.detach().mean(dim=1)
            self.a = self.quantiles * torch.diff(e.detach(), dim=1).mean(dim=1)

        else:

            grad = self.a.unsqueeze(1) * x + self.b.unsqueeze(1) - e.detach()
            self.b -= 2 * self.lr * grad.mean(dim=1)
            self.a -= 2 * self.lr * (grad * x).mean(dim=1)

        l = (e - (self.a.unsqueeze(1) * x + self.b.unsqueeze(1))) ** 2
        return l.mean(dim=1).sum()


class SplineEmbedding(nn.Module):

    def __init__(self, n_features, n_quantiles, emb_dim, n_tables=1, enable=True, init_weights=None, sparse=False):
        super().__init__()

        self.n_features = n_features
        self.emb_dim = emb_dim
        self.enable = enable

        self.n_tables = n_tables
        self.n_emb = (n_quantiles + 2) * n_features
        self.register_buffer('ind_offset', (n_quantiles + 2) * torch.arange(n_features, dtype=torch.int64).unsqueeze(0))
        self.register_buffer("n_quantiles", torch.FloatTensor([n_quantiles]))

        if init_weights is None:
            self.emb = nn.Embedding(self.n_emb * n_tables, emb_dim, sparse=sparse)
        else:

            none_val, base_1, base_0 = init_weights

            q = (torch.arange(n_quantiles + 1) / n_quantiles).view(1, 1, -1, 1)
            weights = base_1.unsqueeze(2) * q + base_0.unsqueeze(2) * (1 - q)

            weights = torch.cat([none_val.unsqueeze(2), weights], dim=2).reshape(-1, emb_dim)

            self.emb = nn.Embedding.from_pretrained(weights, sparse=sparse, freeze=False)

    def forward(self, x, mask, rand_table):

        # if self.enable:
        #     if not self.n_features:
        #         return torch.Tensor(len(x), 0, self.emb_dim, device=x.device, dtype=x.dtype)
        # else:
        #     return torch.zeros(*x.shape, self.emb_dim, device=x.device, dtype=x.dtype)

        th = 1e-3 if x.dtype == torch.float16 else 1e-6
        x = torch.clamp(x, min=th, max=1-th)

        offset = self.ind_offset + self.n_emb * rand_table

        xli = (x * self.n_quantiles).floor().long()
        xl = xli / self.n_quantiles

        xhi = (x * self.n_quantiles + 1).floor().long()
        xh = xhi / self.n_quantiles

        bl = self.emb((xli + 1) * mask + offset)
        bh = self.emb((xhi + 1) * mask + offset)

        delta = 1 / self.n_quantiles
        h = bh / delta * (x - xl).unsqueeze(2) + bl / delta * (xh - x).unsqueeze(2)

        return h


class PID(object):

    def __init__(self, k_p=0.05, k_i=0.005, k_d=0.005, T=20, clip=0.005):
        super().__init__()

        self.k_p = k_p
        self.k_i = k_i
        self.k_d = k_d
        self.T = T
        self.eps_list = []
        self.val_loss_list = []
        self.clip = clip

    def __call__(self, eps, val_loss=None):

        self.eps_list.append(eps)
        self.eps_list = self.eps_list[-self.T:]
        if val_loss is not None:
            self.val_loss_list.append(val_loss)
            self.val_loss_list = self.val_loss_list[-self.T:]
        r = self.k_p * eps + self.k_i * np.sum(self.eps_list)
        if val_loss is None:
            if len(self.eps_list) > 1:
                r += self.k_d * (eps - self.eps_list[-2])
        else:
            if len(self.val_loss_list) > 1:
                r += self.k_d * (val_loss - self.val_loss_list[-2])
        r = self.clip * np.tanh(r / (self.clip + 1e-8))

        return r


class LazyQuantileNorm(nn.Module):

    def __init__(self, quantiles=100, momentum=.001,
                 track_running_stats=True, use_stats_for_train=True, boost=True, predefined=None, scale=True):
        super().__init__()

        quantiles = torch.arange(quantiles) / (quantiles - 1)

        n_quantiles = len(quantiles) if predefined is None else predefined.shape[-1]
        self.register_buffer("n_quantiles", torch.FloatTensor([n_quantiles]))

        boundaries = None if predefined is None else predefined
        self.predefined = False if predefined is None else True

        self.register_buffer("boundaries", boundaries)
        self.register_buffer("quantiles", quantiles)

        self.lr = momentum
        self.boost = boost
        self.scale = scale

        assert (not use_stats_for_train) or (use_stats_for_train and track_running_stats)

        self.track_running_stats = track_running_stats
        self.use_stats_for_train = use_stats_for_train

    def forward(self, x):

        dtype = x.dtype

        if not self.track_running_stats or self.boundaries is None:

            boundaries = torch.quantile(x.float(), self.quantiles.float(), dim=0).transpose(0, 1)
            boundaries = boundaries.type(dtype)

            if self.boundaries is None:
                self.boundaries = boundaries
        else:
            if self.training and not self.predefined:

                q = self.quantiles.view(1, 1, -1)
                b = self.boundaries.unsqueeze(0)
                xv = x.unsqueeze(-1).detach()
                q_th = (q * (xv - b) > (1 - q) * (b - xv)).type(dtype)
                q_grad = (- q * q_th + (1 - q) * (1 - q_th)) * (~torch.isinf(xv)).type(dtype)
                q_grad = q_grad.sum(dim=0)

                if self.boost:
                    q = self.quantiles.unsqueeze(0)
                    factor = (torch.max(1 / (q + 1e-3), 1 / (1 - q + 1e-3))) ** 0.5
                else:
                    factor = 1

                self.boundaries = self.boundaries - self.lr * torch.std(x, dim=0).unsqueeze(-1) * factor * q_grad
                self.boundaries = self.boundaries.sort(dim=1).values


        if (self.training and self.use_stats_for_train) or (
                not self.training and self.track_running_stats) or self.predefined:
            boundaries = self.boundaries

        xq = torch.searchsorted(boundaries, x.transpose(0, 1)).transpose(0, 1)
        if self.scale:
            xq = xq / self.n_quantiles

        return xq


class BetterEmbedding(torch.nn.Module):

    def __init__(self, numerical_indices, categorical_indices, n_quantiles, n_categories, emb_dim,
                 momentum=.001, track_running_stats=True, n_tables=15, initial_mask=1.,
                 k_p=0.05, k_i=0.005, k_d=0.005, T=20, clip=0.005, quantile_resolution=1e-4,
                 use_stats_for_train=True, boost=True, flatten=False, quantile_embedding=True, tokenizer=True,
                 qnorm_flag=False, kaiming_init=False, init_spline_equally=True, sparse=True, spline=True):

        super().__init__()

        self.categorical_indices = categorical_indices
        self.numerical_indices = numerical_indices

        n_feature_num = len(numerical_indices)
        n_features = len(torch.cat([categorical_indices, numerical_indices]))

        n_categories = n_categories + 1
        cat_offset = n_categories.cumsum(0) - n_categories
        self.register_buffer("cat_offset", cat_offset.unsqueeze(0))
        self.register_buffer("null_emb_cat", torch.FloatTensor(1, 0, emb_dim))

        self.flatten = flatten
        self.n_tables = n_tables

        self.n_emb = int(n_categories.sum())

        self.pid = PID(k_p=k_p, k_i=k_i, k_d=k_d, T=T, clip=clip)
        self.br = initial_mask

        if len(categorical_indices):
            self.emb_cat = nn.Embedding(1 + self.n_emb * n_tables, emb_dim, sparse=sparse)
        else:
            self.emb_cat = lambda x: self.null_emb_cat.repeat(len(x), 1, 1)

        if init_spline_equally:

            none_val = torch.randn(n_tables, n_feature_num, emb_dim)
            base_1 = torch.randn(n_tables, n_feature_num, emb_dim)
            base_2 = torch.randn(n_tables, n_feature_num, emb_dim)
            weights = (none_val, base_1, base_2)
        else:
            weights = None

        self.emb_num = None
        if spline:
            self.emb_num = SplineEmbedding(n_feature_num, n_quantiles, emb_dim, n_tables=n_tables,
                                       enable=quantile_embedding, init_weights=weights, sparse=sparse)

        self.llr = LazyLinearRegression(n_quantiles, lr=0.001)
        self.lambda_llr = 0
        if n_quantiles > 1:
            self.pid_llr = PID(k_p=1e-1, k_i=0, k_d=0, T=T, clip=1e-3)
        else:
            self.pid_llr = PID(k_p=1e-4, k_i=0, k_d=0, T=T, clip=0)

        self.qnorm = None
        if qnorm_flag:
            self.qnorm = LazyQuantileNorm(quantiles=int(1 / quantile_resolution), momentum=momentum,
                                      track_running_stats=track_running_stats,
                                      use_stats_for_train=use_stats_for_train, boost=boost)

        self.tokenizer = tokenizer
        if tokenizer:
            self.weight = nn.Parameter(torch.empty((1, n_features, emb_dim)))
            if kaiming_init:
                nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
            else:
                nn.init.normal_(self.weight)
        else:
            self.register_buffer('weight', torch.zeros((n_features, emb_dim)))

        self.bias = nn.Parameter(torch.empty((1, n_features, emb_dim)))
        if kaiming_init:
            nn.init.kaiming_uniform_(self.bias, a=math.sqrt(5))
        else:
            nn.init.normal_(self.bias)

        self.emb_dim = emb_dim

    def step(self, train_loss, val_loss):

        self.br = min(max(0, self.br + self.pid((train_loss - val_loss) / val_loss)), 1)
        self.lambda_llr = min(max(0, self.lambda_llr - self.pid_llr((train_loss - val_loss) / val_loss)), 1)
        logger.info(f"br was changed to {self.br}")
        logger.info(f"lambda_llr was changed to {self.lambda_llr}")

    def get_llr(self):
        return self.llr(self.emb_num.emb.weight) * self.lambda_llr

    def forward(self, x, ensemble=True):

        x_num = x[:, self.numerical_indices]

        if self.qnorm is not None:
            x_num = self.qnorm(x_num)
        x_cat = x[:, self.categorical_indices].long()

        if ensemble:

            bernoulli = torch.distributions.bernoulli.Bernoulli(probs=self.br)
            mask_num = bernoulli.sample(sample_shape=x_num.shape).long().to(x.device)
            mask_cat = bernoulli.sample(sample_shape=x_cat.shape).long().to(x.device)

        else:
            mask_num = 1
            mask_cat = 1

        if self.training:
            rand_table = torch.randint(self.n_tables, size=(1, 1), device=x_cat.device)
        else:
            rand_table = torch.randint(self.n_tables, size=(len(x), 1), device=x_cat.device)

        x_cat = (x_cat + 1) * mask_cat + self.cat_offset + self.n_emb * rand_table

        e_cat = self.emb_cat(x_cat)

        if self.emb_num is not None:
            e_num = self.emb_num(x_num, mask_num, rand_table)
        else:
            e_num = torch.zeros(*x_num.shape, self.emb_dim, device=x_num.device, dtype=x_num.dtype)

        e = torch.cat([e_cat, e_num], dim=1)
        e = e + self.bias

        if self.tokenizer:
            x = torch.cat([torch.zeros_like(x_cat), x_num * mask_num], dim=1)
            y = self.weight * x.unsqueeze(-1)
            e = e + y
        if self.flatten:
            e = e.view(len(e), -1)

        return e


class mySequential(nn.Sequential):
    def forward(self, *input):
        for module in self._modules.values():
            input = module(*input)
        return input


class GBN(torch.nn.Module):
    """
    Ghost Batch Normalization
    https://arxiv.org/abs/1705.08741
    """

    def __init__(self, input_dim, virtual_batch_size=128, momentum=0.01):
        super().__init__()

        self.input_dim = input_dim
        self.virtual_batch_size = virtual_batch_size
        self.bn = nn.BatchNorm1d(self.input_dim, momentum=momentum)

    def forward(self, x):
        chunks = x.chunk(int(np.ceil(x.shape[0] / self.virtual_batch_size)), 0)
        res = [self.bn(x_) for x_ in chunks]

        return torch.cat(res, dim=0)


class Sparsemax(nn.Module):
    """Sparsemax function."""

    def __init__(self, dim=None):
        """Initialize sparsemax activation

        Args:
            dim (int, optional): The dimension over which to apply the sparsemax function.
        """
        super().__init__()

        self.dim = -1 if dim is None else dim

    def forward(self, input):
        """Forward function.
        Args:
            input (torch.Tensor): Input tensor. First dimension should be the batch size
        Returns:
            torch.Tensor: [batch_size x number_of_logits] Output tensor
        """
        # Sparsemax currently only handles 2-dim tensors,
        # so we reshape to a convenient shape and reshape back after sparsemax
        input = input.transpose(0, self.dim)
        original_size = input.size()
        input = input.reshape(input.size(0), -1)
        input = input.transpose(0, 1)
        dim = 1

        number_of_logits = input.size(dim)

        # Translate input by max for numerical stability
        input = input - torch.max(input, dim=dim, keepdim=True)[0].expand_as(input)

        # Sort input in descending order.
        # (NOTE: Can be replaced with linear time selection method described here:
        # http://stanford.edu/~jduchi/projects/DuchiShSiCh08.html)
        zs = torch.sort(input=input, dim=dim, descending=True)[0]
        range = torch.arange(start=1, end=number_of_logits + 1, step=1, device=input.device, dtype=input.dtype).view(1,
                                                                                                                     -1)
        range = range.expand_as(zs)

        # Determine sparsity of projection
        bound = 1 + range * zs
        cumulative_sum_zs = torch.cumsum(zs, dim)
        is_gt = torch.gt(bound, cumulative_sum_zs).type(input.type())
        k = torch.max(is_gt * range, dim, keepdim=True)[0]

        # Compute threshold function
        zs_sparse = is_gt * zs

        # Compute taus
        taus = (torch.sum(zs_sparse, dim, keepdim=True) - 1) / k
        taus = taus.expand_as(input)

        # Sparsemax
        self.output = torch.max(torch.zeros_like(input), input - taus)

        # Reshape back to original shape
        output = self.output
        output = output.transpose(0, 1)
        output = output.reshape(original_size)
        output = output.transpose(0, self.dim)

        return output

    def backward(self, grad_output):
        """Backward function."""
        dim = 1

        nonzeros = torch.ne(self.output, 0)
        sum = torch.sum(grad_output * nonzeros, dim=dim) / torch.sum(nonzeros, dim=dim)
        self.grad_input = nonzeros * (grad_output - sum.expand_as(grad_output))

        return self.grad_input


class BeamEnsemble(torch.nn.Module):

    def __init__(self, net, n_ensembles, optimizer=None):
        super().__init__()

        if isinstance(net, nn.Module):
            ensembles = []
            for _ in range(n_ensembles):
                new_net = copy_network(net)
                reset_network(new_net)
                ensembles.append(new_net)
        else:
            ensembles = [net() for _ in range(n_ensembles)]

        self.ensembles = nn.ModuleList(ensembles)
        self.n_ensembles = n_ensembles

        self.optimizers = None
        if optimizer is not None:
            self.optimizers = self.set_optimizers(optimizer)

        self.optimizer = None
        self.net = None
        self.active_model = None

    def set_optimizers(self, optimizer):

        try:
            self.optimizers = [optimizer(net) for net in self.ensembles]
        except TypeError:
            self.optimizers = [optimizer(net.paramters()) for net in self.ensembles]

        return self.optimizers

    def __len__(self):
        return self.n_ensembles

    def forward(self, x, reduction='mean', index=None):

        if self.training:
            if index is None:
                    index = random.randint(0, self.n_ensembles-1)

            self.net = self.ensembles[index]
            if self.optimizers is not None:
                self.optimizer = self.optimizers[index]
            self.active_model = index

            y = self.net(x)

        else:
            self.optimizer = None
            self.net = None
            self.active_model = None

            y = [net(x) for net in self.ensembles]
            if reduction == 'mean':
                y = torch.stack(y).mean(dim=0)
            elif reduction == 'none':
                y = torch.stack(y)
            else:
                raise NotImplementedError

        return y


class FeatureHasher(object):

    def __init__(self, num_embeddings, embedding_dim, n_classes=None, distribution='uniform', seed=None, device='cpu'):

        manual_seed = seed is not None
        with torch.random.fork_rng(devices=[device], enabled=manual_seed):

            if manual_seed:
                torch.random.manual_seed(seed)

            rand_func = torch.rand if distribution == 'uniform' else torch.randn
            self.weight = rand_func(num_embeddings, embedding_dim, device=device)

            if n_classes is not None:
                self.weight = (self.weight * n_classes).long()

    def __call__(self, x):
        return self.weight[x]


def beam_weights_initializer(net, black_list=None, white_list=None, zero_bias=True,
                             method=None, gain=None, nonlinearity='relu', method_argv=None,
                             nonlinearity_argv=None, method_linear=None, method_linear_argv=None,
                             method_conv=None, method_conv_argv=None, method_embedding=None, method_embedding_argv=None,
                             temperature=1):
    """

    @param net:
    @param black_list:
    @param white_list:
    @param zero_bias:
    @param method: [kaiming_uniform, kaiming_normal, xavier_uniform, xavier_normal, orthogonal, trunc_normal, sparse]
    @param gain:
    @param nonlinearity:
    @param nonlinearity_argv:
    @return:
    """

    def bias_init(m):
        if zero_bias:
            nn.init.zeros_(m.bias)
        else:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(m.weight)
            if fan_in != 0:
                bound = 1 / math.sqrt(fan_in)
                nn.init.uniform_(m.bias, -bound, bound)

    if gain is None:
        if nonlinearity_argv is None:
            nonlinearity_argv = {}
        gain = nn.init.calculate_gain(nonlinearity, **nonlinearity_argv)

    # bert initializes embedding of 768 with .02 std and pytorch initializes with 1. stdout
    # so by default we use something in the middle which is also more reasonable for embedding of ~256 dimensions.
    if method_embedding is None:
        method_embedding = partial(nn.init.normal_, mean=0., std=.2)
    else:
        if method_embedding_argv is None:
            method_embedding_argv = {}
        method_embedding = partial(getattr(nn.init, f"{method_embedding}_"), **method_embedding_argv)

    if method is not None:
        if method_argv is None:
            method_argv = {}
        method = partial(getattr(nn.init, f"{method}_"), **method_argv)

    if method_conv is not None:
        if method_conv_argv is None:
            method_conv_argv = {}
        method_conv = partial(getattr(nn.init, f"{method_conv}_"), **method_conv_argv)
    elif method is not None:
        method_conv = method
    else:
        method_conv = partial(nn.init.xavier_uniform_, gain=gain)

    if method_linear is not None:
        if method_linear_argv is None:
            method_linear_argv = {}
        method_linear = partial(getattr(nn.init, f"{method_conv}_"), **method_linear_argv)
    elif method is not None:
        method_linear = method
    else:
        method_linear = partial(nn.init.xavier_uniform_, gain=gain)

    if black_list is None:
        black_list = []
    if white_list is None:
        # white_list = [n for n, p in net.named_parameters()]
        white_list = ['']

    def valid_param(n, suffix=None):
        if suffix is not None:
            n = f"{n}.{suffix}"
        return any([nw in n for nw in white_list]) and not any([nb in n for nb in black_list])

    for n, m in net.named_modules():
        if len(list(m.children())) or not valid_param(n):
            continue

        if 'Norm' in str(type(m)):
            if valid_param(n, suffix='weight') and hasattr(m, 'weight'):
                nn.init.ones_(m.weight)
            if valid_param(n, suffix='bias') and hasattr(m, 'bias'):
                nn.init.ones_(m.bias)
        elif 'Embedding' in str(type(m)):
            if valid_param(n, suffix='weight') and hasattr(m, 'weight'):
                method_embedding(m.weight)

        elif 'Linear' in str(type(m)):
            if valid_param(n, suffix='weight') and hasattr(m, 'weight'):
                fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(m.weight)
                if fan_in > 1 and fan_out > 1:
                    method_linear(m.weight)
                else:
                    with torch.no_grad():
                        m.weight.normal_()
                        m.weight.data = torch.exp(m.weight.data / (temperature * math.sqrt(torch.numel(m.weight))))
                        m.weight.data = m.weight.data / m.weight.data.sum()

            if valid_param(n, suffix='bias') and hasattr(m, 'bias'):
                bias_init(m)
        elif 'Conv' in str(type(m)):
            if valid_param(n, suffix='weight') and hasattr(m, 'weight'):
                fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(m.weight)
                if fan_in > 1 and fan_out > 1:
                    method_conv(m.weight)
                else:
                    with torch.no_grad():
                        m.weight.normal_()
                        m.weight.data = torch.exp(m.weight.data / (temperature * math.sqrt(torch.numel(m.weight))))
                        m.weight.data = m.weight.data / m.weight.data.sum()
            if valid_param(n, suffix='bias') and hasattr(m, 'bias'):
                bias_init(m)
        else:
            if len(list(m.parameters())) > 0:
                logger.warning(f"Beam weight initializer does not support layer type: {n}")


def reset_network(net):
    prev_hash = {n: hash_tensor(p) for n, p in net.named_parameters()}
    for n, m in net.named_modules():
        if hasattr(m, 'reset_parameters'):
            m.reset_parameters()
    for n, p in net.named_parameters():
        if prev_hash[n] == hash_tensor(p):
            logger.warning(f"Parameter {n} was not reset. Check if its nn.Module supports .reset_parameters()")


def free_network_params(*nets):
    for net in nets:
        for p in net.parameters():
            p.requires_grad = True


def freeze_network_params(*nets):
    for net in nets:
        for p in net.parameters():
            p.requires_grad = False


def copy_network(net):
    return copy.deepcopy(net)


def soft_target_update(source, target, tau):
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(target_param.data * tau + param.data * (1.0 - tau))


def target_copy(source, target):
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(param.data)


def reset_networks_and_optimizers(networks=None, optimizers=None):
    if networks is not None:
        net_iter = networks.keys() if isinstance(networks, dict) else range(len(networks))
        for i in net_iter:
            for n, m in networks[i].named_modules():
                if hasattr(m, 'reset_parameters'):
                    m.reset_parameters()

    if optimizers is not None:
        opt_iter = optimizers.keys() if isinstance(optimizers, dict) else range(len(optimizers))
        for i in opt_iter:
            opt = optimizers[i]

            if type(opt) is BeamOptimizer:
                opt.reset()
            else:
                opt.state = defaultdict(dict)


import torch
from torch import nn
from collections import defaultdict
import math
from functools import partial
from .utils import logger

class MultipleScheduler(object):

    def __init__(self, multiple_optimizer, scheduler, *argc, **argv):

        self.schedulers = {}
        self.multiple_optimizer = multiple_optimizer

        for op in multiple_optimizer.optimizers.keys():
            self.schedulers[op] = scheduler(multiple_optimizer.optimizers[op], *argc, **argv)

    def step(self, *argc, **argv):
        for op in self.multiple_optimizer.optimizers.keys():
            self.schedulers[op].step(*argc, **argv)

    def state_dict(self):
        return {k: sch.state_dict() for k, sch in self.schedulers.items()}

    def load_load_state_dict(self, state):
        for k, sch in self.schedulers.items():
            if k in state:
                sch.load_state_dict(state[k])
            else:
                logger.error(f"Missing scheduler key from state_dict: {k}")


class BeamScheduler(object):

    def __init__(self, optimizer, total_steps=None, epochs=None, steps_per_epochs=None,
                 warmup=5, method='one_cycle', step_type='epoch',
                 pct_start=0.3, anneal_strategy='cos', cycle_momentum=True, base_momentum=0.85, start_factor=0.3,
                 max_momentum=0.95, div_factor=25.0, eta_min=1e-6, factor=math.sqrt(.1), patience=None,
                 threshold=0.0001, T_0=10, T_mult=1, threshold_mode='rel', cooldown=0, min_lr=1e-6):

        self.method = method
        self.epoch = 0
        self.warmup_scheduler = None
        self.warmup = warmup
        self.optimizer = optimizer
        self.last_lr = None
        self.last_momentum = None

        if method == 'one_cycle':
            self.step_type = 'iteration'
        else:
            self.step_type = step_type

        self.total_steps = self.get_total_steps(total_steps=total_steps, epochs=epochs,
                                                steps_per_epochs=steps_per_epochs)

        if method == 'one_cycle':

            max_lr = optimizer.param_groups[0]['lr']
            if self.total_steps is None:
                scheduler = partial(torch.optim.lr_scheduler.OneCycleLR, optimizer=optimizer, max_lr=max_lr,
                                                                pct_start=pct_start, anneal_strategy=anneal_strategy,
                                                                cycle_momentum=cycle_momentum, base_momentum=base_momentum,
                                                                max_momentum=max_momentum, div_factor=div_factor)
            else:
                scheduler = torch.optim.lr_scheduler.OneCycleLR(optimizer, max_lr, total_steps=self.total_steps,
                                                            pct_start=pct_start, anneal_strategy=anneal_strategy,
                                                            cycle_momentum=cycle_momentum, base_momentum=base_momentum,
                                                            max_momentum=max_momentum, div_factor=div_factor)
        else:

            if self.warmup is not None and self.warmup > 0:
                self.warmup_scheduler = torch.optim.lr_scheduler.LinearLR(optimizer, start_factor=start_factor,
                                                                          total_iters=warmup)
                if self.total_steps is not None:
                    self.total_steps = self.total_steps - self.warmup

            if method == 'reduce_on_plateau':

                if patience is None and self.total_steps is not None:
                    patience = self.patience_heuristics(self.total_steps)

                if patience is not None:
                    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='max', factor=factor,
                                                                           patience=patience, threshold=threshold,
                                                                           threshold_mode=threshold_mode,
                                                                           cooldown=cooldown, min_lr=min_lr)
                else:
                    scheduler = partial(torch.optim.lr_scheduler.ReduceLROnPlateau, optimizer=optimizer, mode='max',
                                        factor=factor, threshold=threshold,
                                        threshold_mode=threshold_mode, cooldown=cooldown, min_lr=min_lr)

            elif method == 'cosine_annealing':

                if self.total_steps is None:
                    scheduler = partial(torch.optim.lr_scheduler.CosineAnnealingLR, optimizer=optimizer,
                                        eta_min=eta_min)
                else:
                    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, self.total_steps, eta_min=eta_min)

            elif method == 'cawr':

                scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=T_0, T_mult=T_mult,
                                                                                 eta_min=eta_min)

            else:
                scheduler = None

        self.scheduler = scheduler

    def get_total_steps(self, total_steps=None, epochs=None, steps_per_epochs=None):
        if epochs is not None and self.step_type == 'epoch':
            total_steps = epochs
        elif epochs is not None and steps_per_epochs is not None:
            total_steps = epochs * steps_per_epochs
        return total_steps

    @staticmethod
    def patience_heuristics(total_steps):

        # return 2 * int(np.log2(total_steps / 12.5))

        if total_steps > 400:
            return 10
        if total_steps > 200:
            return 8
        if total_steps > 100:
            return 6
        if total_steps > 50:
            return 4
        return 2

    def update_total_steps(self, total_steps=None, epochs=None, steps_per_epochs=None, initial_state=None):

        self.total_steps = self.get_total_steps(total_steps=total_steps, epochs=epochs,
                                                steps_per_epochs=steps_per_epochs)

        if self.warmup_scheduler is not None:
            self.total_steps = self.total_steps - self.warmup
        if type(self.scheduler) is partial:
            if self.method == 'one_cycle':
                self.scheduler = self.scheduler(total_steps=self.total_steps)
            elif self.method == 'cosine_annealing':
                self.scheduler = self.scheduler(T_max=self.total_steps)
            elif self.method == 'reduce_on_plateau':
                self.scheduler = self.scheduler(patience=self.patience_heuristics(self.total_steps))
            else:
                raise NotImplementedError(f"Method: {self.method} is still unsupported")

            if initial_state is not None:
                self.scheduler.load_state_dict(initial_state)

    def get_current_state(self):

        lr = self.optimizer.param_groups[0]['lr']
        if self.method in ['one_cycle']:
            if self.scheduler.use_beta1:
                momentum = self.optimizer.param_groups[0]['betas'][0]
            else:
                momentum = self.optimizer.param_groups[0]['momentum']
        else:
            momentum = None

        return {'lr': lr, 'momentum': momentum}

    def state_dict(self):
        return {'epoch': 0,
                'warmup_scheduler': None if self.warmup_scheduler is None else self.warmup_scheduler.state_dict(),
                'scheduler': self.scheduler.state_dict()}

    def load_state_dict(self, state):
        self.epoch = state['epoch']
        self.scheduler.load_state_dict(state['scheduler'])
        if self.warmup_scheduler is not None:
            self.warmup_scheduler.load_state_dict(state['warmup_scheduler'])

    def step(self, objective=None, step_type=None):

        if step_type != self.step_type and step_type is not None:
            return
        if self.warmup_scheduler is not None and self.epoch < self.warmup:
            self.warmup_scheduler.step()
        else:
            if self.method == 'reduce_on_plateau':
                self.scheduler.step(objective)
            else:
                if self.scheduler is not None:
                    self.scheduler.step()

        self.epoch = self.epoch + 1
        self.get_current_state()


class BeamOptimizer(object):

    def __init__(self, net, dense_args=None, clip=0, accumulate=1, amp=False,
                 sparse_args=None, dense_optimizer='AdamW', sparse_optimizer='SparseAdam'):

        sparse_optimizer = getattr(torch.optim, sparse_optimizer)
        dense_optimizer = getattr(torch.optim, dense_optimizer)

        if dense_args is None:
            dense_args = {'lr': 1e-3, 'eps': 1e-4}
        if sparse_args is None:
            sparse_args = {'lr': 1e-2, 'eps': 1e-4}

        self.clip = clip
        self.accumulate = accumulate
        self.iteration = 0
        self.amp = amp
        self.autocast_device = next(net.parameters()).device.type
        self.scaler = torch.cuda.amp.GradScaler() if self.amp else None

        self.optimizers = {}

        sparse_parameters = []
        dense_parameters = []

        for nm, m in net.named_modules(remove_duplicate=True):
            is_sparse = BeamOptimizer.check_sparse(m)
            if is_sparse:
                for n, p in m.named_parameters(recurse=False):
                    if not any([p is pi for pi in sparse_parameters]):
                        sparse_parameters.append(p)
            else:
                for n, p in m.named_parameters(recurse=False):
                    if not any([p is pi for pi in dense_parameters]):
                        dense_parameters.append(p)

        if len(dense_parameters) > 0:
            self.optimizers['dense'] = dense_optimizer(dense_parameters, **dense_args)

        if len(sparse_parameters) > 0:
            self.optimizers['sparse'] = sparse_optimizer(sparse_parameters, **sparse_args)

        for k, o in self.optimizers.items():
            setattr(self, k, o)

    @staticmethod
    def prototype(dense_args=None, clip=0, accumulate=1, amp=False,
                  sparse_args=None, dense_optimizer='AdamW', sparse_optimizer='SparseAdam'):
        return partial(BeamOptimizer, dense_args=dense_args, clip=clip, accumulate=accumulate, amp=amp,
                       sparse_args=sparse_args, dense_optimizer=dense_optimizer, sparse_optimizer=sparse_optimizer)

    @staticmethod
    def check_sparse(m):
        return (isinstance(m, nn.Embedding) or isinstance(m, nn.EmbeddingBag)) and m.sparse

    def set_scheduler(self, scheduler, *argc, **argv):
        return MultipleScheduler(self, scheduler, *argc, **argv)

    def reset(self):
        self.iteration = 0
        for op in self.optimizers.values():
            op.state = defaultdict(dict)

        self.zero_grad(set_to_none=True)
        self.scaler = torch.cuda.amp.GradScaler() if self.amp else None

    def zero_grad(self, set_to_none=True):
        for op in self.optimizers.values():
            op.zero_grad(set_to_none=set_to_none)

    def apply(self, loss, set_to_none=True, gradient=None, retain_graph=None, create_graph=False, inputs=None):

        with torch.autocast(self.autocast_device, enabled=False):
            self.iteration += 1

            if self.amp:
                self.scaler.scale(loss).backward(gradient=gradient, retain_graph=retain_graph,
                                                 create_graph=create_graph, inputs=inputs)
            else:
                loss.backward(gradient=gradient, retain_graph=retain_graph,
                              create_graph=create_graph, inputs=inputs)

            if self.clip > 0:
                for op in self.optimizers.values():
                    if self.amp:
                        self.scaler.unscale_(op)
                    for pg in op.param_groups:
                        torch.nn.utils.clip_grad_norm_(iter(pg['params']), self.clip)

            if not (self.iteration % self.accumulate):
                self.step()
                self.zero_grad(set_to_none=set_to_none)

    def step(self):

        for op in self.optimizers.values():
            if self.amp:
                self.scaler.step(op)
            else:
                op.step()

        if self.amp:
            self.scaler.update()

    def state_dict(self):
        state_dict = {k: op.state_dict() for k, op in self.optimizers.items()}
        state_dict['scaler'] = self.scaler.state_dict() if self.scaler is not None else None
        return state_dict

    def load_state_dict(self, state_dict, state_only=False):

        for k, op in self.optimizers.items():

            if state_only:
                state_dict[k]['param_groups'] = op.state_dict()['param_groups']

            op.load_state_dict(state_dict[k])

        if self.scaler is not None and 'scaler' in state_dict.keys():
            self.scaler.load_state_dict(state_dict["scaler"])
